(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
Require Import ZArith.
Require Import ZArith_ext.

(** OmegaZ : extension of omega to handle goals/hypos that contains Z_of_nat *)

Lemma Z0_mech: Z0 = Z_of_nat 0. Proof. auto. Qed.
  
Lemma Z_of_nat_inj': forall x y, x = y -> Z_of_nat x = Z_of_nat y.
Proof. intros x y H; rewrite H; reflexivity. Qed.
    
Lemma nat_of_Z: forall z, (z >= 0)%Z -> exists n, z = Z_of_nat n.
Proof.
intros x H.
apply Z_of_nat_complete.
apply Zge_le.
assumption.
Qed.

Ltac mult_dist ome :=
  (rewrite mult_plus_distr_l; (mult_dist || idtac)) || 
  (rewrite mult_plus_distr_r; (mult_dist || idtac)) || 
  (rewrite mult_minus_distr_r; [(mult_dist || idtac) | OmegaZ ome]) || 
  (rewrite mult_minus_distr_l; [(mult_dist || idtac) | OmegaZ ome]) || 
  idtac
with 
  Z_to_nat x:= 
  assert (Z_to_natA1 : Zpos x = Z_of_nat (nat_of_P x)); 
    [auto | rewrite Z_to_natA1; clear Z_to_natA1; unfold nat_of_P; simpl (Pmult_nat x 1); idtac]
with 
  Supr_Z ome t := 
  match t with 
    | Z0 => rewrite Z0_mech
    | Zpos ?x => Z_to_nat x
    | Zplus ?x1 ?x2 => 
      ((Supr_Z ome x1 || idtac); (Supr_Z ome x2 || idtac); rewrite <- inj_plus) || 
      idtac
    | Zminus ?x1 ?x2 => 
      ((Supr_Z ome x1 || idtac); (Supr_Z ome x2 || idtac); rewrite <- inj_minus1; [idtac | ome]) || 
      idtac
    | Zmult ?x1 ?x2 => 
      ((Supr_Z ome x1 || idtac); (Supr_Z ome x2 || idtac); rewrite <- inj_mult) || 
      idtac
    | Z_of_nat ?x => idtac
    | ?E => 
      assert (Supr_Z'A1: (E >= 0)%Z); 
        [ome | 
         generalize (nat_of_Z _ Supr_Z'A1); clear Supr_Z'A1; let X := fresh in intro X; 
         let x := fresh in (let y := fresh in (destruct X as [x y]; rewrite y))] || 
      idtac
    | _ => idtac
  end
with
  Supr_Z_eq_left ome :=
  match goal with
    | |- (?x1 = ?x2)%Z => Supr_Z ome x1
    | |- (?x1 <= ?x2)%Z => Supr_Z ome x1
    | |- (?x1 >= ?x2)%Z => Supr_Z ome x1
    | |- (?x1 > ?x2)%Z => Supr_Z ome x1
    | |- (?x1 < ?x2)%Z => Supr_Z ome x1
    | |- _ => idtac
  end
with
  Supr_Z_eq_right ome :=
  match goal with
    | |- (?x1 = ?x2)%Z => Supr_Z ome x2
    | |- (?x1 <= ?x2)%Z => Supr_Z ome x2
    | |- (?x1 >= ?x2)%Z => Supr_Z ome x2
    | |- (?x1 > ?x2)%Z => Supr_Z ome x2
    | |- (?x1 < ?x2)%Z => Supr_Z ome x2
    | |- _ => idtac
  end
with
  Supr_Z_eq :=
  match goal with
    | |- (Z_of_nat ?x1 = Z_of_nat ?x2)%Z => apply Z_of_nat_inj'
    | |- (Z_of_nat ?x1 <= Z_of_nat ?x2)%Z => apply inj_le
    | |- (Z_of_nat ?x1 < Z_of_nat ?x2)%Z => apply inj_lt
    | |- (Z_of_nat ?x1 >= Z_of_nat ?x2)%Z => apply inj_ge
    | |- (Z_of_nat ?x1 < Z_of_nat ?x2)%Z => apply inj_gt
    | |- _ => idtac
  end
with
  Z_to_nat_Hyp x id:= 
  assert (Z_to_natA1 : Zpos x = Z_of_nat (nat_of_P x)); 
    [auto | 
     rewrite Z_to_natA1 in id; clear Z_to_natA1; unfold nat_of_P in id; simpl (Pmult_nat x 1) in id; idtac ]
with
  Supr_Z_Hyp ome t id := 
  match t with 
    | Z0 => rewrite Z0_mech in id
    | Zpos ?x => Z_to_nat_Hyp x id
    | Zplus ?x1 ?x2 => (Supr_Z_Hyp ome x1 id || idtac) ; (Supr_Z_Hyp ome x2 id || idtac); rewrite <- inj_plus in id
    | Zminus ?x1 ?x2 => (Supr_Z_Hyp ome x1 id || idtac) ; (Supr_Z_Hyp ome x2 id || idtac); rewrite <- inj_minus1 in id; [idtac | ome]
    | Zmult ?x1 ?x2 => (Supr_Z_Hyp ome x1 id || idtac) ; (Supr_Z_Hyp ome x2 id || idtac); rewrite <- inj_mult in id
    | Z_of_nat ?x => idtac
    | ?E => 
      assert (Supr_Z'A1: (E >= 0)%Z); 
        [ome | 
         generalize (nat_of_Z _ Supr_Z'A1); clear Supr_Z'A1; let X := fresh in intro X; 
         let x := fresh in (let y := fresh in (destruct X as [x y]; rewrite y in id))] || 
      idtac
    | _ => idtac
  end
  
with
  Supr_Z_eq_hyp_left ome :=
  match goal with
    | id: (?x1 = ?x2)%Z |- _ => Supr_Z_Hyp ome x1 id
    | id: (?x1 <= ?x2)%Z |- _ => Supr_Z_Hyp ome x1 id
    | id: (?x1 >= ?x2)%Z |- _ => Supr_Z_Hyp ome x1 id
    | id: (?x1 > ?x2)%Z |- _ => Supr_Z_Hyp ome x1 id
    | id: (?x1 < ?x2)%Z |- _ => Supr_Z_Hyp ome x1 id
    | |- _ => idtac
  end
with
  Supr_Z_eq_hyp_right ome :=
  match goal with
    | id: (?x1 = ?x2)%Z |- _ => Supr_Z_Hyp ome x2 id
    | id: (?x1 <= ?x2)%Z |- _ => Supr_Z_Hyp ome x2 id
    | id: (?x1 >= ?x2)%Z |- _ => Supr_Z_Hyp ome x2 id
    | id: (?x1 > ?x2)%Z |- _ => Supr_Z_Hyp ome x2 id
    | id: (?x1 < ?x2)%Z |- _ => Supr_Z_Hyp ome x2 id
    | |- _ => idtac
  end
with Supr_Z_eq_Hyp := 
  match goal with
    | id: (?x1 = ?x2)%Z |- _ => generalize (Z_of_nat_inj _ _ id); intros
    | id: (?x1 > ?x2)%Z |- _ => generalize ((proj2 (Nat2Z.inj_gt _ _)) id); intros
    | id: (?x1 < ?x2)%Z |- _ => generalize ((proj2 (Nat2Z.inj_lt _ _)) id); intros
    | id: (?x1 >= ?x2)%Z |- _ => generalize ((proj2 (Nat2Z.inj_ge _ _)) id); intros
    | id: (?x1 <= ?x2)%Z |- _ => generalize ((proj2 (Nat2Z.inj_le _ _)) id); intros
    | |- _ => idtac
  end
with
  OmegaZ ome := 
  ((Supr_Z_eq_hyp_left ome; Supr_Z_eq_hyp_right ome; Supr_Z_eq_Hyp); (Supr_Z_eq_left ome; Supr_Z_eq_right ome; Supr_Z_eq) || idtac); 
  mult_dist ome; (ome || tauto).

Goal forall x y,
  ((Z_of_nat x) + 4%Z - 2%Z > (Z_of_nat (y +4)))%Z ->
  ((Z_of_nat (x + 2)) + 2%Z >= (Z_of_nat y) + 6%Z)%Z.
Proof.
intros.
OmegaZ omega.
Qed.

Ltac omegaz_inj_plus_hyp :=
  match goal with
    | id : context [Z_of_nat (plus _ _)] |- _ => rewrite inj_plus in id
    | id : context [Z_of_nat (ssrnat.addn _ _)] |- _ => rewrite inj_plus in id
  end.

Ltac omegaz_inj_plus :=
  match goal with
    | |- context [Z_of_nat (plus _ _)] => rewrite inj_plus
    | |- context [Z_of_nat (ssrnat.addn _ _)] => rewrite inj_plus
  end.

Ltac omegaz_inj_minus_hyp :=
  match goal with
    | id : context [Z_of_nat (minus _ _)] |- _ => rewrite inj_minus1 in id
    | id : context [Z_of_nat (ssrnat.subn _ _)] |- _ => rewrite inj_minus1 in id
  end.

Ltac omegaz_inj_minus :=
  match goal with
    | |- context [Z_of_nat (minus _ _)] => rewrite inj_minus1
    | |- context [Z_of_nat (ssrnat.subn _ _)] => rewrite inj_minus1
  end.

Ltac omegaz_inj_mult_hyp :=
  match goal with
    | id : context [Z_of_nat (mult _ _)] |- _ => rewrite inj_mult in id
    | id : context [Z_of_nat (ssrnat.muln _ _)] |- _ => rewrite inj_mult in id
  end.

Ltac omegaz_inj_mult :=
  match goal with
    | |- context [Z_of_nat (mult _ _)] => rewrite inj_mult
    | |- context [Z_of_nat (ssrnat.muln _ _)] => rewrite inj_mult
  end.

Ltac simpl_Z_of_nat_cst_hyp :=
  match goal with
    | id : context [Z_of_nat O] |- _ => simpl (Z_of_nat O) in id
    | id : context [Z_of_nat (S ?x)] |- _ => simpl (Z_of_nat (S x)) in id
  end.

From mathcomp Require Import ssreflect ssrbool.

Ltac simpl_Z_of_nat_cst :=
  match goal with
    | |- context [Z_of_nat O] => simpl (Z_of_nat O)
    | |- context [Z_of_nat 1] => rewrite [Z_of_nat 1]/=
    | |- context [Z_of_nat 2] => rewrite [Z_of_nat 2]/=
    | |- context [Z_of_nat 3] => rewrite [Z_of_nat 3]/=
    | |- context [Z_of_nat 4] => rewrite [Z_of_nat 4]/=
    | |- context [Z_of_nat (S ?x)] => rewrite (Z_S x)
  end.

Ltac Zabs_nat_tac_hyp :=
  match goal with
    | id : context [Zabs_nat (Z_of_nat ?n)] |- _ => rewrite (Zabs_nat_Z_of_nat n) in id
    | id : context [Zabs_nat 1%Z] |- _ => rewrite (_ : Zabs_nat 1%Z = 1%nat) in id; last by done
  end.

(* TODO: could be removed (and the tactic just above also?) *)
Ltac Zabs_nat_tac :=
  match goal with
    | |- context [Zabs_nat (Z_of_nat ?n)] => rewrite (Zabs_nat_Z_of_nat n)
    | |- context [Zabs_nat 1%Z] => rewrite (_ : Zabs_nat 1%Z = 1%nat); last by done
  end.

Ltac omegaz_hyp :=
  repeat omegaz_inj_plus_hyp ;
  repeat omegaz_inj_minus_hyp ;
  repeat omegaz_inj_mult_hyp ;
  repeat simpl_Z_of_nat_cst_hyp.
  
Ltac omegaz_goal :=
  repeat omegaz_inj_plus ;
  repeat omegaz_inj_minus ;
  repeat omegaz_inj_mult ;
  repeat simpl_Z_of_nat_cst.

Ltac omegaz' ome :=
  repeat omegaz_hyp ;
  repeat omegaz_goal ;
  ome.

Ltac omegaz := omegaz' omega.

Goal forall x y,
  ((Z_of_nat x) + 4%Z - 2%Z > (Z_of_nat (y +4)))%Z ->
  ((Z_of_nat (x + 2)) + 2%Z >= (Z_of_nat y) + 6%Z)%Z.
Proof.
intros.
omegaz.
Qed.
