(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrbool eqtype ssrnat seq.
Require Import ZArith_ext seq_ext ssrnat_ext.

(** Definition of C-like strings and some properties.

Preliminary. Used in a Seplog example that illustrated
buffer-overflows. *)

Definition string' (lst : list nat) := ~ List.In O lst.

Definition string (lst : list Z) := exists lst',
  string' lst' /\ lst = (map (fun x => Z_of_nat x) lst') ++ (0::nil).

Lemma string_nil : ~ string nil.
Proof.
move=> [x [H H1]].
have H0 : (1 <= size (@nil Z))%nat by rewrite H1 size_cat /=; ssromega.
inversion H0.
Qed.

Lemma string'_sub : forall lst, string' lst -> string' (List.tail lst).
Proof.
induction lst; intros; auto.
simpl.
red in H.
red.
simpl in H.
tauto.
Qed.

Lemma string_sub : forall lst, (2 <= size lst)%nat -> string lst -> string (List.tail lst).
Proof.
induction lst; intros; auto.
simpl.
red in H0.
inversion_clear H0.
inversion_clear H1.
red.
exists (List.tail x).
split.
apply string'_sub; auto.
destruct x.
simpl in H2.
rewrite H2 in H.
inversion H.
inversion H2.
simpl.
simpl in H2.
injection H2; auto.
Qed.

Lemma string_sup : forall lst, string lst ->
  forall lst', string' lst' -> string (map (fun x => Z_of_nat x) lst' ++ lst).
Proof.
induction lst'.
simpl.
auto.
simpl.
intros.
generalize (string'_sub _ H0); intro.
simpl in H1.
generalize (IHlst' H1); intro.
red in H2.
inversion_clear H2 as [lst''].
red.
exists (a::lst'').
inversion_clear H3.
split.
red.
red in H2.
red in H0.
simpl in H0.
simpl.
tauto.
rewrite /= H4 //.
Qed.

Lemma string_last : forall lst, string lst -> forall i, nth (-1) lst i = 0 -> 
  (i = size lst - 1)%nat.
Proof.
induction lst; intros.
  generalize string_nil; tauto.
simpl.
rewrite subSS subn0.
destruct lst.
  red in H.
  inversion_clear H.
  inversion_clear H1.
  destruct x.
    simpl in H2.
    injection H2; clear H2; intro.
    subst a.
    simpl in H0.
    destruct i; auto.
    destruct i; discriminate.
  assert (2 <= size (a::nil))%nat.
    rewrite H2.
    rewrite size_cat.
    simpl.
    rewrite size_map.
    ssromega.
  done.
inversion H.
inversion H1.
destruct i.
  simpl in H0.
  subst a.
  red in H.
  inversion_clear H.
  inversion_clear H0.
  destruct x.
    simpl in H1.
    discriminate.
  simpl in H3.
  injection H3; clear H3; intros.
  assert (n=O).
    omega.
  subst n.
  red in H2.
  simpl in H2.
  tauto.
assert (string (z :: lst)).
  cutrewrite (z::lst = List.tail (a::z::lst)); [idtac | auto].
  apply string_sub.
    simpl.
    ssromega.
  assumption.
lapply (IHlst H4 i).
  intros.
  rewrite H5.
  simpl.
  by rewrite subSS subn0.
rewrite //= in H0 *.
Qed.

Lemma string_hd_ge0: forall a l, string (a::l) -> a >= 0.
Proof.
intros.
inversion_clear H.
inversion_clear H0.
destruct x.
injection H1.
intros.
omega.
simpl in H1.
injection H1.
intros.
omega.
Qed.

Lemma string_last' : forall i lst, string lst -> (i = size lst - 1)%nat ->
  nth (-1) lst i = 0.
Proof.
move=> i lst [lst' [H H2]] Hi.
subst lst.
rewrite size_cat /= in Hi.
by rewrite nth_cat size_map {1}Hi size_map addn1 subn1 ltnn Hi size_map addn1 subn1 /= subnn.
Qed.

Lemma string_last'' : forall i lst, string lst -> (i < size lst)%nat -> 
  nth (-1) lst i >= 0.
Proof.
induction i; intros.
  red in H.
  inversion_clear H as [lst'].
inversion_clear H1.
destruct lst'.
simpl in H2.
rewrite H2.
simpl.
omega.
rewrite H2.
simpl.
red in H.
simpl in H.
omega.
inversion_clear H as [lst'].
inversion_clear H1.
destruct lst.
destruct lst'; discriminate.
simpl.
apply IHi.
destruct lst'; try discriminate.
simpl in H2.
destruct lst; try discriminate.
simpl in H0.
inversion H0.
inversion H3.
simpl in H2.
injection H2; clear H2; intros.
subst z.
rewrite H1.
red.
exists lst'.
split; auto.
generalize (string'_sub _ H); intro.
auto.
simpl in H0.
ssromega.
Qed.
