(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
Ltac typeof X := type of X.

From mathcomp Require Import ssreflect ssrbool eqtype.
Require ProofIrrelevance.

Lemma not_Some_is_None {A : Type} (v : option A) :
  ~ (exists k, v = Some k) -> v = None.
Proof. destruct v => // H; exfalso. apply H; by exists a. Qed.

Lemma None_is_not_Some {A : Type} (v : option A) :
  v = None -> ~ (exists k, v = Some k).
Proof. move=> ->; by case. Qed.

Notation "'|_' x '_|'" := (Some x) (at level 55, right associativity) : core_scope.

Notation "'|_' s , h '_|'" := (Some (s, h)) (at level 55, right associativity) : core_scope.

Lemma optionT_dec : forall {A : Type} (a : option A),
  sigT (fun l => a = Some l) + {a = None}.
Proof.
move=> A [].
left; by exists a.
by right.
Defined.

Definition option_dec {A : eqType} (a : option A) :
  sigT (fun l => a = Some l) + {a = None} := 
match a with
| Some a' => inleft (@existT A (fun l => |_ a' _| = |_ l _|) a' Logic.eq_refl)
| None => inright Logic.eq_refl
end.

Lemma option_decT : forall {A : Type} (H : forall (x y : A), {x = y} + {x <> y}) (a : option A),
  sigT (fun l => a = Some l) + {a = None}.
Proof.
move=> A H [].
left; by exists a.
by right.
Defined.

(*Lemma eqtype_is_comparable : forall (A : eqType) (x y : A), {x = y} + {x <> y}.
Proof.
move=> A x y.
case => sort0 m T x y.
case: m x y => op m x y.
case: (m x y).
by left.
by right.
Qed.*)

Lemma eq_comparable_eq (T : eqType) (x : T) : 
 eq_comparable x x = left (x <> x) (Logic.eq_refl x).
Proof.
case: eq_comparable.
move=> H.
f_equal.
by apply eq_irrelevance.
move=> b.
exfalso.
done.
Qed.

Lemma eq_comparable_neq (T : eqType) (x y: T) (H: x <> y):   
 eq_comparable x y = right (x = y) H.
Proof.
case: eq_comparable.
move=> H'.
exfalso.
apply H; apply H'.
move=> H'.
f_equal.
by apply ProofIrrelevance.proof_irrelevance.
Qed.

(*Definition coerce_eqType {A : eqType} {B : A -> Type} {a1 a2 : A} (H: a1 = a2) (x: B a1) : B a2.
rewrite -H.
exact x.
Defined.

Lemma coerce_eqType_eq {A: eqType} {B} {a: A} (H: a = a) (x: B a):
  coerce_eqType H x = x.
rewrite /coerce_eqType.
have Heq: H = (Logic.eq_refl _) by apply /eq_irrelevance.
by subst H.
Qed.*)

Lemma constructive_indefinite_description' :
  forall (A : Type) (P : A -> Prop), { x : A | P x } -> exists x, P x.
Proof. move=> A P [x' Px']; by exists x'. Qed.
