(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq.
Require Import ZArith.
Notation "m ^ n" := (expn m n) : nat_scope. (* TODO: kludge *)
Require Import Znumtheory.
Require Import Bool_ext ssrZ.
Export ZArith.
Export Znumtheory.

Local Close Scope Z_scope.
Local Close Scope positive_scope.

Local Close Scope Z_scope.

Notation "'nat<=Z'" := (Z.abs_nat) (at level 9) : zarith_ext_scope.
Local Open Scope zarith_ext_scope.

Lemma positiveP : Equality.axiom Ndec.Peqb.
Proof.
move=> x y.
apply: (iffP idP) => H.
move: {H}(Ndec.Peqb_Pcompare _ _ H) => H.
apply Pcompare_Eq_eq in H => //.
subst x.
apply: Ndec.Peqb_correct.
Qed.
  
Canonical Structure positive_eqMixin := EqMixin positiveP.
Canonical Structure positive_eqType := Eval hnf in EqType _ positive_eqMixin.

(** Lemmas on Zeq_bool *)

Lemma Zeq_bool_neq' x y : x <> y -> Zeq_bool x y = false.
Proof. move=> H. by apply/negbTE/eqP. Qed.

(** Lemmas on Z *)

Local Open Scope Z_scope.

(* NB: see Coq.ZArith.Zorder.Zlt_neg_0 for the Zneg case *)
Lemma Zlt_pos_0 : forall p, 0 < Zpos p. Proof. by case. Qed.

Lemma Zlt0_pos a b : 0 <= a -> 0 <= b -> a + b != 0 -> 0 < a + b.
Proof. move=> Ha Hb a_b. move/eqP : a_b => ?. omega. Qed.

Lemma Zopp_eq_mult_neg_1_L (n : Z) : -1 * n = - n.
Proof. by rewrite Zmult_comm Zopp_eq_mult_neg_1. Qed.

Lemma Zplus_Zopp a b : a + - b = a - b.
Proof. omega. Qed.

Lemma Znot_lt_le n m : ~ m < n -> n <= m.
Proof. move=> H. apply Znot_gt_le. contradict H. apply Zgt_lt => //. Qed.

Lemma Zle_minus_1 a b : a < b -> a <= b - 1.
Proof. omega. Qed.

Lemma Zminus_le_decr a b : 0 <= b -> a - b <= a.
Proof. move=> Hb. omega. Qed.

Lemma neg_Zlt : forall n, 0 < n -> - n < 0. Proof. by case. Qed.

Lemma Zmult_sign : forall a b, 0 <= a -> b <= 0 -> a * b <= 0.
Proof.
move=> [|a|a] // b Ha Hb.
by case: b Hb.
Qed.

Lemma Zle_scale a b : 0 <= a -> 0 < b -> a <= a * b.
Proof.
move=> Ha Hb.
rewrite (_ : a * b = a + a * (b - 1)); last by ring.
rewrite -{1}(Zplus_0_r a).
apply/Zplus_le_compat_l/Zmult_le_0_compat; omega.
Qed.

Lemma Zle_scale_neg a b : a <=0 -> 0 < b -> a * b <= a.
Proof.
move=> Ha Hb.
rewrite (_ : a * b = a + a * (b - 1)); last by ring.
rewrite -{3}(Zplus_0_r a).
apply Zplus_le_compat_l.
rewrite Zmult_comm.
apply Zmult_sign => //; omega.
Qed.

Lemma Zle_plus_trans a b c : 0 <= b -> a <= c -> a <= b + c.
Proof.
move=> Hb a_c.
rewrite -(Zplus_0_l a).
by apply Zplus_le_compat.
Qed.

Lemma Zle_plus_trans_L a b c : b <= 0 -> a <= c -> a + b <= c.
Proof.
move=> Hb a_c.
rewrite -(Zplus_0_r c).
by apply Zplus_le_compat.
Qed.

Lemma Z_opposite_sign : forall a b, a < 0 -> 0 <= b -> a <> b.
Proof. case=> //= p. by case=> //= q _ []. Qed.

Lemma Zminus_le_compat a b c : a <= c -> a - b <= c - b.
Proof. intros; omega. Qed.

Lemma Zplus_lt_reg_r_not A B C : (~ A + C < B + C)%Z -> (~ A < B)%Z.
Proof. move=> H. contradict H. by apply Zplus_lt_compat_r. Qed.

(* NB: Coq.ZArith.Zorder.Zmult_lt_compat2 *)
Lemma Zlt_Zmult_inv' a b c : 0 <= a -> 0 < b -> 0 <= c -> a * b < c -> a < c.
Proof.
move=> Ha Hb Hc Habc.
apply Zmult_gt_0_lt_reg_r with b; first by omega.
apply Zlt_le_trans with c => //.
apply Zle_scale; omega.
Qed.

Lemma Zlt_Zmult_inv a b c : (a * b < c -> 0 < a -> 0 <= c -> b < c)%Z.
Proof.
move=> H Ha Hc.
case: (Z_lt_le_dec b 0) => Hb.
by apply Zlt_le_trans with Z0.
apply Zlt_Zmult_inv' with a => //.
by rewrite Zmult_comm.
Qed.

Lemma Zlt_Zplus_inv a b c : 0 <= b -> a + b < c -> a < c.
Proof. move=> Hb ab_c. omega. Qed.

Lemma Zmult_lt_reg_l n m p : 0 < p -> p * n < p * m -> n < m.
Proof.
move=> Hp H.
rewrite 2!(Zmult_comm p) in H.
by apply Zmult_lt_reg_r with p.
Qed.

Lemma Zlt_compat a b c d : b < c -> a = b -> c = d -> a < d.
Proof. move=> X Y Z; by subst. Qed.

Lemma Zle_neq_lt n m : n <= m -> n <> m -> n < m.
Proof. move=> *; omega. Qed.

Lemma neg_Zle: forall n : Z, 0 <= n -> - n <= 0.
Proof. by case. Qed.

Lemma Zlt_neg_pos p q : Zneg p < Zpos q.
Proof. move: (Zle_neg_pos p q) => H; by apply Zle_neq_lt. Qed.

Lemma Zlt_opp a b : - b < - a -> a < b.
Proof. move=> b_a; omega. Qed.

(** Lemmas on Z_of_nat *)

Lemma Z_of_nat_Zpos_P_of_succ_nat : forall n, Z_of_nat (n + 1) = Zpos (P_of_succ_nat n).
Proof.
case=> //= n.
by rewrite -P_of_succ_nat_o_nat_of_P_eq_succ nat_of_P_o_P_of_succ_nat_eq_succ -addnE addn1.
Qed.

Lemma Z_of_nat_inj : forall x y, Z_of_nat x = Z_of_nat y -> x = y.
Proof.
elim.
- by case.
- move=> n IH; case=> //= m; case=> H.
  by rewrite -2!nat_of_P_o_P_of_succ_nat_eq_succ H.
Qed.

Lemma Z_of_nat_inj_neq x y : Z_of_nat x <> Z_of_nat y -> x <> y.
Proof. move=> H H'; by apply H; f_equal. Qed.

Lemma Z_S n : Z_of_nat n.+1 = Z_of_nat n + 1.
Proof. by rewrite inj_S. Qed.

Lemma Z_of_nat_neg : forall z, (Z_of_nat z <= 0)%Z -> z = O.
Proof. by case. Qed.

Lemma Z_of_nat_0 : forall a, Z_of_nat a = 0 -> a = O. 
Proof. by case. Qed.

Lemma Z_of_nat_lt0 : forall a, (0 < a)%nat -> 0 < Z_of_nat a.
Proof. case=> //; by inversion 1. Qed.

Lemma Zmult_neg' a b : a < 0 -> (0 < b)%nat -> a * Z_of_nat b < 0.
Proof.
elim: b a => [a Ha | [|b] IHb a Ha _].
- by rewrite ltnn.
- by rewrite Zmult_1_r.
- rewrite Z_S Zmult_plus_distr_r Zmult_1_r -(Zplus_0_r 0).
  apply Zplus_lt_compat => //; by apply IHb.
Qed.

Lemma Zmult_neg a b : a < 0 -> 0 < b -> a * b < 0.
Proof.
move=> Ha Hb.
move/Zlt_le_weak : (Hb) => Hb'.
move/Z_of_nat_complete : Hb' => [[|b'] Hb'].
- rewrite Hb' /= in Hb.
  move: (Zlt_irrefl 0); contradiction.
- rewrite Hb'.
  by apply Zmult_neg'.
Qed.

Lemma Zplus_pos_inv a b c d : a <= c -> b <= d -> a + b = c + d -> a = c /\ b = d.
Proof. move=> H H1 H2. omega. Qed.

Lemma Zplus0_inv a b : 0 <= a -> 0 <= b -> a + b = 0 -> a = 0 /\ b =0.
Proof. 
move=> Ha Hb H. move:(Zplus_pos_inv _ _ _ _ Ha Hb). rewrite H. 
by case/(_ refl_equal) => -> ->. 
Qed.

Lemma Zle_plus_0_inv a b : a + b <= 0 -> 0 <= a -> b <= 0.
Proof. move=> H. omega. Qed.

Lemma Zle_0_mult_inv a b : 0 <= a * b -> (0 <= a /\ 0 <= b) \/ (a <= 0 /\ b <= 0).
Proof. 
move=> H; destruct a.
- destruct b; omega.
- destruct b.
  + omega.
  + by left.
  + move: (Zlt_pos_0 p) => H1.
    move: (Zlt_neg_0 p0) => H2.
    have H3 : Zpos p * Zneg p0 < 0.
      rewrite Zmult_comm.
      by apply Zmult_neg.
    omega.
- destruct b.
  + omega.
  + move: (Zlt_pos_0 p0) => H1.
    move: (Zlt_neg_0 p) => H2.
    have H3 : Zneg p * Zpos p0 < 0.
      by apply Zmult_neg.
    omega.
  + by right.
Qed.

(* NB: see also Zmult_lt_0_reg_r *)
Lemma Zle_mult_0_inv a b : a * b <= 0 -> 0 < a -> b <= 0.
Proof. 
move=> H Ha.
case: (Z_lt_le_dec 0 b) => // Hb.
move: (Zmult_lt_0_compat a b Ha Hb).
by move/Zlt_not_le.
Qed.

Lemma Zlt_mult_0_inv a b :  a * b < 0 -> (a < 0 /\ 0 < b) \/ (0 < a /\ b < 0).
Proof. 
move=> H; destruct a.
- rewrite Zmult_0_l in H.
  by apply Zlt_irrefl in H.
- destruct b.
  + omega.
  + by left.
  + move: (Zlt_pos_0 p) => H1.
    move: (Zlt_neg_0 p0) => H2.
    have H3 : Zpos p * Zneg p0 < 0.
      rewrite Zmult_comm.
      by apply Zmult_neg.
    omega.
- destruct b.
  + omega.
  + move: (Zlt_pos_0 p0) => H1.
    move: (Zlt_neg_0 p) => H2.
    have H3 : Zneg p * Zpos p0 < 0.
      by apply Zmult_neg.
    omega.
  + by right.
Qed.

Lemma Zlt_mult_interval_inv b k a : (0 < k)%Z -> (- b <= k * a < b)%Z -> (- b <= a < b)%Z.
Proof.
move=> Hk [H1 H2].
split.
  case: (Z_lt_ge_dec 0 a) => ?.
    apply Zle_trans with Z0; omega.
  eapply Zle_trans; first by apply H1.
  rewrite -{2}(Zmult_1_l a).
  apply Z.mul_le_mono_nonpos_r.
  by apply Zge_le.
  omega.
case: (Z_lt_ge_dec 0 a) => ?.
  eapply Zle_lt_trans; last by apply H2.
  rewrite -{1}(Zmult_1_l a).
  apply Z.mul_le_mono_pos_r => //.
  omega.
omega.
Qed.

(** Lemmas on Zabs/Zabs_nat *)

Lemma Zabs_pos : forall z, 0 <= Zabs z. Proof. by case. Qed.

Lemma Zabs_Z_of_nat : forall n, Zabs (Z_of_nat n) = Z_of_nat n.
Proof. by case. Qed.

Lemma Zsgn_Zabs_1 : forall z, z <> 0 -> Zsgn (Zabs z) = 1.
Proof. by case. Qed.

Lemma Zabs_Zsgn_1 : forall z, z <> 0 -> Zabs (Zsgn z) = 1.
Proof. by case. Qed.

Lemma Zabs_nat_0_inv : forall a, Zabs_nat a = O -> a = 0.
Proof. case=> //= p H; move: (lt_O_nat_of_P p); rewrite H; by inversion 1. Qed.

Lemma Zabs_nat_mult : forall a b, Zabs_nat (a * b) = (Zabs_nat a * Zabs_nat b)%nat.
Proof. move=> [|p|p] [|q|q] //=; by rewrite nat_of_P_mult_morphism. Qed.

Lemma Zabs_nat_Zneg_inj : forall a b, a < 0 -> b < 0 -> Zabs_nat a = Zabs_nat b -> a = b.
Proof. case=> // p; case => // q _ _ /= H. apply nat_of_P_inj in H; by subst. Qed.

Lemma Zabs_nat_Zpos_inj : forall a b, 0 < a -> 0 < b -> Zabs_nat a = Zabs_nat b -> a = b.
Proof. case=> // p; case => // q _ _ /= H. apply nat_of_P_inj in H; by subst. Qed.

Lemma Zabs_nat_Zpos_Z0_inj a b : 0 <= a -> 0 <= b -> Zabs_nat a = Zabs_nat b -> a = b.
Proof.
move=> Ha Hb.
apply Zle_lt_or_eq in Ha.
apply Zle_lt_or_eq in Hb.
case:Ha=>Ha.
- case:Hb=>Hb.
  + by apply Zabs_nat_Zpos_inj.
  + subst => /=; by apply Zabs_nat_0_inv.
- subst; case:Hb=>Hb.
  + move=> /= H. symmetry; by apply Zabs_nat_0_inv.
  + by subst.
Qed.

Lemma Zabs_nat_Zopp : forall z, Zabs_nat (- z) = Zabs_nat z. Proof. by case. Qed.

Lemma Z_of_nat_Zabs_nat : forall z, 0 <= z -> Z_of_nat (Zabs_nat z) = z.
Proof. case => //= p Hp; by rewrite Zpos_eq_Z_of_nat_o_nat_of_P. Qed.

Lemma Z_of_nat_Zabs_nat_neg : forall (z : Z), z <= 0 -> Z_of_nat (Zabs_nat z) = - z.
Proof. case=> // p _; by rewrite - Zpos_eq_Z_of_nat_o_nat_of_P. Qed.

Lemma S_Zabs_nat : forall n, 0 <= n -> (Zabs_nat n).+1 = Zabs_nat (n + 1).
Proof.
case => //; rewrite /= => p _.
by rewrite nat_of_P_plus_morphism // plus_comm.
Qed.
Lemma O_lt_Zabs_nat z : 0 < z -> (O < Zabs_nat z)%nat.
Proof. 
move=> Hz. rewrite (_ : O = Zabs_nat 0) //. 
apply/ltP/Zabs_nat_lt; split; auto; by apply Zle_refl. 
Qed.

Lemma Zabs_nat_Zsgn : forall z, z <> 0 -> Zabs_nat (Zsgn z) = 1%nat.
Proof. by case. Qed.

Lemma Z_of_nat_Zsgn : forall l, Z_of_nat l = Zsgn (Z_of_nat l) * Z_of_nat l.
Proof. by case. Qed.

Lemma Zsgn_pos0 : forall z, 0 <= z <-> 0 <= Zsgn z.
Proof. by case. Qed.

Lemma Zsgn_neg0 : forall z, z <= 0 <-> Zsgn z <= 0.
Proof. by case. Qed.

Lemma Zodd_Zsgn : forall z, z <> 0 -> Zodd (Zsgn z).
Proof. by case. Qed.

Lemma Zabs_mod z b : 0 < b -> z mod b = 0 -> Zabs z mod b = 0.
Proof.
move=> Hb Hz.
case: (Zabs_dec z) => [<- // | X].
move/Z_mod_zero_opp_full : Hz => Hz.
by rewrite X Zopp_involutive in Hz.
Qed.

Lemma Zabs_lt b z : - b < z < b -> Zabs z < b.
Proof.
move=> [H1 H2].
case: (Z_le_gt_dec 0 z) => Hz.
by rewrite Zabs_eq.
rewrite Zabs_non_eq; last by apply Zlt_le_weak, Zgt_lt.
omega.
Qed.

Lemma Zabs_le b z : (- b <= z <= b)%Z -> (Zabs z <= b)%Z.
Proof.
move=> [H1 H2].
case: (Z_le_gt_dec 0 z) => Hz.
by rewrite Zabs_eq.
rewrite Zabs_non_eq; last by apply Zlt_le_weak, Zgt_lt.
omega.
Qed.

Lemma Zlt_abs b z : Zabs z < b -> - b < z < b.
Proof.
move=> H.
case: (Z_le_gt_dec z 0) => Hz.
rewrite Zabs_non_eq // in H.
omega.
rewrite Zabs_eq // in H; last by omega.
split; last by done.
omega.
Qed.

Lemma Zodd_abs : forall z, Zodd z -> Zodd (Zabs z). Proof. by case. Qed.

Lemma Zlt_0_Zdiv a b : 0 < b -> b <= a -> 0 < a / b.
Proof.
move/Zlt_gt => Hb Hba.
rewrite (_ : a = b * 1 + (a - b)); last by omega. 
rewrite Zmult_comm Z_div_plus_full_l; last by omega.
apply Zlt_le_trans with 1; first by auto.
have Htmp : 0 <= (a - b) / b.
  apply Z_div_pos => //.
  omega.
omega.
Qed.

Lemma Z_div_neg a b : b > 0 -> a <= 0 -> a / b <= 0.
Proof.
move=> Hb Ha.
apply Zdiv_le_upper_bound => //.
by apply Zgt_lt.
Qed.

Lemma Z_div_neg_lt a b : b > 0 -> a < 0 -> a / b < 0.
Proof.
intros.
apply Zdiv_lt_upper_bound => //.
by apply Zgt_lt.
Qed.

Lemma Zdiv_le_pos : forall z b, 0 <= z -> 0 < b -> z / b <= z.
Proof.
move=> z b Hz Hb.
apply Zdiv_le_upper_bound => //.
by apply Zle_scale.
Qed.

Lemma Zdiv_le_neg : forall z b, z <= 0 -> 0 < b -> z <= z / b.
Proof.
move=> z b Hz Hb.
apply Zdiv_le_lower_bound => //.
by apply Zle_scale_neg.
Qed.

Lemma poly_eq0_inv : forall a b k, 0 <= k -> -k < b < k ->
  a * k + b = 0 -> a = 0 /\ b = 0.
Proof.
move=> a b k H H0 H1.
apply Zle_lt_or_eq in H.
case: H => H; last first.
subst k; omega.
have {H1}H1 : a * k = - b by omega.
case: (Z_lt_ge_dec a 0) => ?.
- have H2 :  b > 0.
    suff ? : -b < 0 by omega.
    rewrite -H1.
    by apply Zmult_neg.
  have H5 : a * k <= - k.
    cutrewrite (-k = (-1)*k); last by omega.
    apply Zmult_le_compat_r; omega.
  rewrite H1 in H5.
  have ? : k <= b by omega.
  omega.
- case: (Z_eq_dec a 0) => H2.
  + split; first by assumption.
    subst a; omega.
  + have H3 : a > 0 by omega.
    have H4 : b < 0.
      suff ? : 0 < -b by omega.
      rewrite -H1.
      apply Zmult_lt_0_compat; omega.
    have H5 : k <= a * k.
      cutrewrite (k = 1 * k); last by omega.
      apply Zmult_le_compat; omega.
    rewrite H1 in H5.
    have ? : -k >= b by omega.
    have [//] : False by omega.
Qed.

Lemma poly_eq_inv : forall a b a' b' k,
  0 <= a /\ 0 <= b < k /\ 0 <= a' /\ 0 <= b' < k ->
  a * k + b = a' * k + b' -> a = a' /\ b = b'.
Proof.
move=> a b a' b' k H H0.
have {H0}H0 : (a - a') * k + (b - b') = 0.
  rewrite Zmult_minus_distr_r; omega.
have H1 : (0 < k) by omega.
have H2 : (-k < b - b' < k) by omega.
move: (poly_eq0_inv _ _ _ (Zlt_le_weak _ _ H1) H2 H0) => ?; omega.
Qed.

Lemma poly_Zlt1_inv : forall b a k,
  0 <= b -> 0 <= a -> 0 <= k ->
  a * k + b < k -> a = 0.
Proof.
move=> b a k Hb Ha Hk H.
destruct a => //.
move=> {Ha}.
have H' : 1 * k <= Zpos p * k.
  apply Zmult_le_compat_r => //.
  by case: p H.
have Habs : k <= b + Zpos p * k by omega.
omega.
Qed.

Lemma poly_Zlt1_inv_new : forall b a k : Z, 0 <= b -> 0 <= a -> 0 <= k -> Zabs (a * k + b) < k -> a = 0.
Proof.
move=> b a k Hb Ha Hk.
rewrite Zabs_eq //; last first.
apply Zplus_le_0_compat => //.
by apply Zmult_le_0_compat.
by apply poly_Zlt1_inv.
Qed.

Lemma poly_Zlt : forall a a' b b' k,
  0 <= a -> 0 <= a' -> 0 <= b -> 0 <= b' -> 0 <= k -> 
  a < k ->
  a >= a' ->
  b < b' ->
  a + k * b < a' + k * b'.
Proof.
move=> a a' b b' k Ha Ha' Hb Hb' Hk Hak Haa' Hbb'.
have [b'' Hb''] : exists b'', b' = 1 + b'' by exists (b'-1); omega.
subst b'.
rewrite Zmult_plus_distr_r Zmult_1_r Zplus_assoc.
apply Zplus_lt_le_compat.
omega.
apply Zmult_le_compat_l => //.
omega.
Qed.

Lemma Zlt_bool_false : forall a b, b >= a -> Zlt_bool b a = false.
Proof. move=> a b. rewrite /Zge /Zlt_bool. by destruct (b ?= a). Qed.

Lemma Zlt_bool_trans : forall a b c, Zlt_bool a b -> Zlt_bool b c -> Zlt_bool a c.
Proof.
move=> a b c.
move/ZltP => Hab.
move/ZltP => Hbc.
apply/ZltP.
by apply Zlt_trans with b.
Qed.

(*Lemma Zge_bool_false : forall a b, Zge_bool a b = false -> a < b.
Proof. move=> a b. rewrite /Zlt /Zge_bool. by destruct (a ?= b). Qed.

Lemma Zgt_bool_false: forall a b, Zgt_bool a b = false -> a <= b.
Proof. move=> a b. rewrite /Zle /Zgt_bool. by destruct (a ?= b). Qed.*)

Lemma ZsgnK : forall z, Zsgn (Zsgn z) = Zsgn z.
Proof. by case. Qed.

Lemma Zsgn_0_le_inv : forall z, 0 <= Zsgn z -> Zsgn z = 0 \/ Zsgn z = 1.
Proof.
case=> [_|p /= _|p /= Hp].
by left.
by right.
by move/ZleP : Hp => /=.
Qed.

Definition bZsgn z := if Zeq_bool z 0 then 1%Z else (-1)%Z.

Lemma Zmod_le : forall a b, 0 <= b -> 0 <= a mod b.
Proof.
move=> a; case.
- move=> _; rewrite Zmod_0_r //.
- move=> p H.
  have X : Zpos p > 0 by done.
  move: (Z_mod_lt a (Zpos p) X).
  by tauto.
- move=> p H.
  move: (Zlt_neg_0 p) => H'; by omega.
Qed.

(** Zpower *)
(** this function is extensionally the same as Coq.ZArith.Zpower.Zpower_nat but it is defined in a more direct way,
    there is already a function Zpower in Coq.ZArith.Zpow_def but with a different type (the exponent can be negative) *)
Fixpoint Zpower (b : Z) (e : nat) { struct e } : Z :=
  match e with
    | O => 1
    | e'.+1 => b * Zpower b e'
  end.

Notation "b ^^ e" := (Zpower b e) (at level 30, right associativity) : zarith_ext_scope.

Lemma ZpowerS : forall k n, k ^^ n.+1 = k * k ^^ n. 
Proof. by rewrite //=. Qed.

Lemma ZpowerD b : forall n m, b ^^ (n + m) = b ^^ n * b ^^ m.
Proof.
elim => [m|n IH m].
by rewrite [Zmult]lock /= -lock Zmult_1_l.
rewrite addSn 2!ZpowerS IH; ring.
Qed.

(*Lemma Zpower_is_exp n m : 2 ^^ (n + m) = 2 ^^ n * 2 ^^ m.
Proof. by apply ZpowerD. Qed.*)

Lemma Zpower_0 : forall n, 0 < 2 ^^ n.
Proof. elim => // n IH; rewrite ZpowerS; omega. Qed.

Lemma Zpower_0' n : 0 <= 2 ^^ n.
Proof. move: (Zpower_0 n) =>X ; omega. Qed.

Lemma Zpower_1 n : 1 <= 2 ^^ n.
Proof. move: (Zpower_0 n) =>X ; omega. Qed.

Lemma Zpower_2_lt : forall n m, (n < m)%nat -> 2 ^^ n < 2 ^^ m.
Proof.
elim => [ [|m _] | n IH [ // |m] ].
- by rewrite ltnn.
- apply Zlt_le_trans with 2 => //.
  rewrite ZpowerS.
  move: (Zpower_1 m) => ?; by omega.
- rewrite ltnS => /IH.
  rewrite 2!ZpowerS => ?; by omega.
Qed.

Lemma Zpower_2_le n m : (2 ^^ n <=? 2 ^^ m) = (n <= m)%nat.
Proof.
apply/esym.
apply/idP/idP.
  elim : n m.
  - move=> m _ /=; by apply/ZleP/Zpower_1.
  - move=> n IH [|m] //.
    rewrite ltnS => /IH.
    rewrite 2!ZpowerS => ?.
    by rewrite Zleq_pmul2l.
move=> H.
rewrite leqNgt; apply/negP => n_m.
move/ZleP : H.
by apply/Zlt_not_le/Zpower_2_lt.
Qed.

Lemma Zpower_2_inv n m : 2 ^^ n.+1 < 2 ^^ m.+1 -> (n < m)%nat.
Proof.
move=> H.
rewrite ltnNge; apply/negP => n_m.
move: H.
apply/Zle_not_lt/ZleP.
by rewrite Zpower_2_le.
Qed.

Lemma Zpower_plus n : 2 ^^ n + 2 ^^ n = 2 ^^ n.+1.
Proof. rewrite ZpowerS; by ring. Qed.

Lemma Zpower_b_square z : z * z = z ^^ 2.
Proof. move=> /=. rewrite Zmult_1_r //. Qed.

Lemma Zpower_unit : forall n, 1 ^^ n = 1.
Proof. elim=> // n; rewrite ZpowerS => ->. done. Qed.

Lemma ZpowerM : forall m n z, (z ^^ n) ^^ m = z ^^ (n * m).
Proof.
elim => [m z /= | n IHn m z].
- by rewrite muln0.
- by rewrite ZpowerS mulnC ZpowerD IHn mulnC.
Qed.

Lemma Zpower_exp_mod : forall m n z modu,
  (z ^^ n mod modu) ^^ m mod modu = z ^^ (n * m) mod modu.
Proof.
elim=> [m z modu /= | n IHn m z modu].
- by rewrite muln0.
- by rewrite ZpowerS mulnC ZpowerD -Zmult_mod_idemp_r IHn -Zmult_mod mulnC.
Qed.

Lemma power_Zpower b n : Z_of_nat (b ^ n) = Z_of_nat b ^^ n.
Proof. elim : n => // n0 IH; by rewrite expnS ZpowerS inj_mult IH. Qed.

Lemma Zabs_power : forall (a : Z) (b : nat), 0 <= a -> Zabs (a ^^ b) = a ^^ b.
Proof.
move=> a.
elim=> // b IH Ha /=.
rewrite Zabs_Zmult Zabs_eq // IH //.
Qed.

Lemma power_Zpower2 : forall n, Zabs_nat (2 ^^ n - 1) = (2 ^ n - 1)%nat.
Proof.
elim => // n IH.
rewrite ZpowerS expnS.
apply Z_of_nat_inj.
rewrite Z_of_nat_Zabs_nat; last first.
  move: (Zpower_0 n) => ?; omega.
rewrite inj_minus1; last first.
  rewrite -{1}(mult_1_l 1).
  apply mult_le_compat; first by auto.
  apply/leP; by rewrite expn_gt0.
rewrite inj_mult.
apply Znat.inj_eq in IH.
rewrite Z_of_nat_Zabs_nat in IH; last first.
  move: (Zpower_0 n) => ?; omega.
rewrite inj_minus1 in IH; last by apply/leP; rewrite expn_gt0.
simpl (Z_of_nat 2).
rewrite power_Zpower in IH *.
simpl Z_of_nat in IH.
simpl Z_of_nat; omega.
Qed.

Lemma Zpower_mod b : forall (n : nat), O <> n -> b ^^ n mod b = 0.
Proof. by move=> [|n] //= _; rewrite Zmult_comm Z_mod_mult. Qed.

Lemma Zpower_div b : forall n, 0 < b -> O <> n -> b ^^ n / b = Zpower b (n - 1).
Proof. move=> [|n] //= Hb _. rewrite Zmult_comm subSS subn0 Z_div_mult_full //. omega. Qed.

Lemma Zpower_nat_Zpower : forall n, Zpower_nat 2 n = 2 ^^ n.
Proof. by elim => // n; rewrite /Zpower_nat ZpowerS => <-. Qed.

Lemma id_rem_minus a b : (a - b) ^^ 2 = a ^^ 2 + b ^^ 2 - 2 * a * b.
Proof. 
rewrite -Zpower_b_square Zmult_minus_distr_r 2!Zmult_minus_distr_l 2!Zpower_b_square; ring. 
Qed.

Lemma id_rem_plus a b : (a + b) ^^ 2 = a ^^ 2 + b ^^ 2 + 2 * a * b.
Proof. 
rewrite -Zpower_b_square Zmult_plus_distr_r 2!Zmult_plus_distr_l 2!Zpower_b_square; ring. 
Qed.

Lemma Zsquare_0_inv z : z ^^ 2 = 0 -> z = 0. Proof. by case : z. Qed.

Lemma Zsquare_pos z : 0 <= z ^^ 2. Proof. by case : z. Qed.

Lemma Zpower_Zmult : forall n a b, (a * b) ^^ n = a ^^ n * b ^^ n.
Proof. elim=> // n IH a b /=. rewrite !IH; ring. Qed.

Lemma Zmod_2_Zeven : forall z, z mod 2 = 0 -> Zeven z.
Proof.
move=> z.
move/Zmod_divide => H.
lapply H; last by done.
move=> {H}H.
destruct H.
rewrite H.
by apply Zeven_mult_Zeven_r.
Qed.

Lemma not_Zmod_2_Zodd : forall z, z mod 2 <> 0 -> Zodd z.
Proof.
move=> z Hz.
case: (Zeven_odd_dec z) => //.
case/Zeven_ex=> z' Hz'; subst z.
by rewrite Zmult_comm Z_mod_mult in Hz.
Qed.

Lemma Zeven_2 : forall m, Zeven (2 * m). Proof. by case. Qed.

Lemma Zodd_1 : forall m, Zodd (2 * m + 1). Proof. elim => //. by case. Qed.

Lemma Zodd_opp : forall z, Zodd z -> Zodd (- z). Proof. by case. Qed.

Lemma Zeven_opp : forall z, Zeven z -> Zeven (- z). Proof. by case. Qed.

Lemma Zeven_opp_inv : forall z, Zeven (- z) -> Zeven z. Proof. by case. Qed.

Lemma Zodd_Zdivide_2 : forall m, Zodd m -> ~ (2 | m ).
Proof.
move=> m Hm [k Hk]; subst m.
apply Zodd_not_Zeven in Hm.
apply Hm.
by rewrite Zmult_comm; apply Zeven_2.
Qed.

Lemma ZevenP z : ssrbool.reflect (Zeven z) (Z.even z).
Proof.
apply: (iffP idP). 
destruct z => //=; by destruct p.
destruct z => //=; by destruct p.
Qed.

Implicit Arguments ZevenP [z].
Prenex Implicits ZevenP.

Definition Zodd_bool z := ~~ Z.even z.

Lemma Zodd_mult_inv : forall a b, Zodd (a * b) -> Zodd a /\ Zodd b.
Proof.
move=> a b H; case : (Zeven_odd_dec a) => Ha.
by move/Zeven_not_Zodd : (Zeven_mult_Zeven_l a b Ha).
case: (Zeven_odd_dec b) => [ Hb | // ].
move/Zeven_not_Zodd : (Zeven_mult_Zeven_l b a Hb).
by rewrite Zmult_comm.
Qed.

Lemma ZoddP : forall z, Zodd z <-> Zodd_bool z.
Proof. rewrite /Zodd_bool; case => //=; by case=> [p|p|]. Qed.

Lemma Zeven_Zmod_2 z : Zeven z -> z mod 2 = 0.
Proof. case/Zeven_ex => z' ->; by rewrite Zmult_comm Z_mod_mult. Qed.

Lemma Zodd_Zmod_2 z : Zodd z -> z mod 2 = 1.
Proof.
case/Zodd_ex => z' ->.
rewrite Zplus_comm Zmult_comm; by apply Z_mod_plus_full.
Qed.

Lemma Zeven_plus_inv : forall a b, Zeven (a + b) ->
  (Zeven a /\ Zeven b) \/ (Zodd a /\ Zodd b).
Proof.
move=> a b H.
case: (Zeven_odd_dec a) => Ha.
case: (Zeven_odd_dec b) => Hb.
tauto.
move/Zodd_not_Zeven: (Zodd_plus_Zeven _ _ Hb Ha).
by rewrite Zplus_comm.
case: (Zeven_odd_dec b) => Hb; last by tauto.
by move/Zodd_not_Zeven : (Zodd_plus_Zeven _ _ Ha Hb).
Qed.

Lemma Zmult_2_Zeven a b : a = 2 * b -> Zeven a.
Proof. move=> ->;  by apply Zeven_2. Qed.

Lemma Zeven_mult_inv a b : Zeven (a * b) -> Zeven a \/ Zeven b.
Proof.
case: (Zeven_odd_dec a) => X; first by auto.
case: (Zeven_odd_dec b) => Y; first by auto.
by move/Zodd_not_Zeven : (Zodd_mult_Zodd _ _ X Y).
Qed.

Lemma Zodd_plus_inv : forall a b, Zodd (a + b) -> (Zodd a /\ Zeven b) \/ (Zeven a /\ Zodd b).
Proof.
move=> a b H.
case: (Zeven_odd_dec a) => Ha; case: (Zeven_odd_dec b) => Hb.
by move/Zeven_not_Zodd : (Zeven_plus_Zeven _ _ Ha Hb).
tauto.
tauto.
by move/Zeven_not_Zodd: (Zodd_plus_Zodd _ _ Ha Hb).
Qed.

Lemma Zodd_bool_Zplus a b : Z.even b -> Zodd_bool (a + b) = Zodd_bool a.
Proof.
move/ZevenP => H.
move H' : (Zodd_bool (a + b)) => c.
destruct c.
- move/ZoddP in H'.
  symmetry.
  apply/ZoddP.
  case/Zodd_plus_inv : H'; first by tauto.
  case=> _.
  by move/Zeven_not_Zodd.
- move/negP : H' => H'.
  symmetry.
  apply/negP.
  contradict H'.
  contradict H'.
  move/ZevenP in H'.
  apply/ZevenP.
  case/Zeven_plus_inv : H' => H'; first by tauto.
  case : H' => _.
  by move/Zodd_not_Zeven.
Qed.

Lemma Zeven_bool_S z : Z.even (z + 1) = ~~ Z.even z.
Proof.
case/boolP : (Z.even z) => Z.
by rewrite Z.even_add /= Z.
by rewrite Z.even_add /= (negbTE Z).
Qed.

Lemma Zeven_power : forall m, Zeven (2 ^^ m.+1).
Proof. elim=> // n H. rewrite ZpowerS (Zeven_div2 _ H). by apply Zeven_2. Qed.

Lemma Zeven_power' : forall m, (0 < m)%nat -> Zeven (2 ^^ m).
Proof. case => // => n _; by apply Zeven_power. Qed.

Lemma Zodd_bool_Zplus_Zpower n (z : Z) b k : (0 < n)%nat -> (0 < k)%nat ->
  Zodd_bool (z + (2 ^^ (k * n)) * b) = Zodd_bool z.
Proof.
move=> Hn Hk.
rewrite Zodd_bool_Zplus //.
apply/ZevenP/Zeven_mult_Zeven_l.
destruct n; first by inversion Hn.
destruct k; first by inversion Hk.
apply/Zeven_power'; by rewrite muln_gt0.
Qed.

Lemma Zpower_Zdivide : forall n d1 d2, d1 >= 0 -> 2 ^^ n = d1 * d2 ->
  exists p, exists q, d1 = 2 ^^ p /\ d2 = 2 ^^ q.
Proof.
elim => [d1 d2 Hd1 /= H | n IH d1 d2 Hd1].
- symmetry in H.
  case: (Zmult_1_inversion_l _ _ H) => H1.
  + subst d1.
    rewrite Zmult_1_l in H.
    subst d2.
    by exists O, O.
  + subst d1.
    suff : ~ (-1 >= 0) by contradiction.
    omega.
- rewrite ZpowerS Zmult_comm => H.
  symmetry in H.
  move: (Zdivide_intro _ _ _ H) => H'.
  case: {H'}(prime_mult _ prime_2 _ _ H') => H'.
  - inversion H' as [kd1 Hkd1]; clear H'.
    have X : 2 ^^ n = kd1 * d2.
      rewrite Hkd1 in H.
      rewrite (Zmult_comm kd1) (Zmult_comm (2 ^^ n)) -Zmult_assoc in H.
      by apply Zmult_reg_l in H.
    apply IH in X; last by omega.
    case: X => [p [q [H1 H2]]].
    exists p.+1, q; by rewrite ZpowerS Hkd1 H1 Zmult_comm.
  - inversion H' as [kd2 Hkd2]; clear H'.
    have X : 2 ^^ n = d1 * kd2.
      rewrite Hkd2 Zmult_assoc in H.
      by apply Zmult_reg_r in H.
    apply IH in X; last by omega.
    case: X => [p [q [H1 H2]]].
    exists p, q.+1; by rewrite ZpowerS Hkd2 H2 Zmult_comm.
Qed.

Definition Zmax_seq l := foldl (fun e a => Z.max e a) 0 l.

Lemma Zis_gcd_eq : forall n m, 0 <= n -> Zis_gcd n n m -> n = m \/ n = - m.
Proof.
move=> n m H [H1 [q [H2]]].
move/(_ n (Zdivide_refl _) (Zdivide_refl _)) => X.
by apply Zdivide_antisym.
Qed.

Lemma Zis_gcd_Zpower_odd : forall n m, Zodd m -> Zis_gcd (2 ^^ n) m 1.
Proof.
move=> n m Hm; constructor.
by apply Zone_divide.
by apply Zone_divide.
move=> d [dpower Hdpower] [dm Hdm].
have [k Hk] : exists k, 1 = k * d.
  have [X|X] : dpower >= 0 \/ dpower < 0 by omega.
  - case: (Zpower_Zdivide _ _ _ X Hdpower) => p [q [H1 H2]].
    case: (leqP q 0) => [|Hq].
    + rewrite leqn0 => /eqP Hq; subst q.
      rewrite /= in H2.
      subst d.
      by exists 1.
    + have Habsurd : (2 | m).
        subst d.
        destruct q.
        * by inversion Hq.
        * exists (dm * 2 ^^ q).
          rewrite Hdm ZpowerS; ring.
      by apply Zodd_Zdivide_2 in Hm.
  - have Y : d < 0.
      destruct d => //.
      * rewrite Zmult_0_r in Hdpower.
        move: (Zpower_0 n) => ?; omega.
      * move: (Zpower_0 n) => Y.
        rewrite Hdpower in Y.
        apply Zmult_gt_0_lt_0_reg_r in Y => //.
        omega.
    have Hdpower' : 2 ^^ n = -dpower * (-d) by rewrite Zmult_opp_opp.
    have X' : (-dpower) >= 0 by omega.
    case: (Zpower_Zdivide _ _ _ X' Hdpower') => p [q [H1 H2]].
    case: (leqP q 0) => [|Hq].
    + rewrite leqn0 => /eqP Hq; subst q.
      rewrite /= in H2.
      by exists (-1).
      have Habsurd : (2 | m).
        have Hdm' : m = (-dm) * (-d).
          rewrite Zmult_opp_opp //.
        rewrite H2 in Hdm'.
        destruct q. 
        * by inversion Hq.
        * exists (-(dm * 2 ^^ q)).
          rewrite Hdm' ZpowerS; ring.
      by apply Zodd_Zdivide_2 in Hm.
    by exists k.
Qed.

Lemma Zis_gcd_even a b g : Zeven a -> Zeven b ->
  Zis_gcd a b (2 * g) -> Zis_gcd (a / 2) (b / 2) g.
Proof.
case/Zeven_ex => qa Ha; subst a.
case/Zeven_ex => qb Hb; subst b => Hgcd.
rewrite (Zmult_comm 2 qa) (Zmult_comm 2 qb) !Z_div_mult //.
case: Hgcd; move=> [a Ha] [b Hb] Hgcd.
rewrite (Zmult_comm a) -Zmult_assoc in Ha.
apply Zmult_reg_l in Ha => //.
rewrite (Zmult_comm b) -Zmult_assoc in Hb.
apply Zmult_reg_l in Hb => //.
apply Zis_gcd_intro.
- rewrite Ha; by apply Zdivide_factor_r.
- rewrite Hb; by apply Zdivide_factor_r.
- move=> x [qqa Hqa] [qqb Hqb]; subst qa qb.
  have [qq Hqq] : (2 * x | 2 * g) by apply Hgcd; apply Zmult_divide_compat_l; eapply Zdivide_intro; eauto.
  rewrite (Zmult_comm qq) -Zmult_assoc in Hqq.
  apply Zmult_reg_l in Hqq => //.
  apply Zdivide_intro with qq.
  by rewrite Zmult_comm.
Qed.

(* NB : lemma Zis_gcd_even_odd in the standard library is for Zpos/Zneg *)
Lemma Zis_gcd_even_odd_new a b g :
  Zis_gcd a (2 * b + 1) g -> Zis_gcd (2 * a) (2 * b + 1) g.
Proof.
case=> Ha Hb Hgcd.
apply Zis_gcd_intro => //.
by apply Zdivide_mult_r.
move=> x Hxa Hxb.
apply Hgcd => //.
have : rel_prime x 2.
  apply bezout_rel_prime.
  case: Hxb => xb Hxb.
  apply Bezout_intro with xb (-b).
  rewrite -Hxb; ring.
by apply Gauss.
Qed.

Lemma Zgcd_0_R : forall a, 0 <= a -> Zgcd a 0 = a.
Proof. move=> a Ha. apply Zis_gcd_gcd => //. by apply Zis_gcd_0. Qed.

Lemma Zgcd_same : forall a, 0 <= a -> Zgcd a a = a.
Proof. move=> a Ha. apply Zis_gcd_gcd => //. by apply Zis_gcd_refl. Qed.

Lemma Zgcd_sym : forall a b, Zgcd a b = Zgcd b a.
Proof.
move=> a b.
apply Zis_gcd_gcd; by [apply Zgcd_is_pos | apply Zis_gcd_sym, Zgcd_is_gcd].
Qed.

Lemma Zgcd_opp a b : Zgcd a (-b) = Zgcd a b.
Proof.
move: (Zis_gcd_minus a (-b) (Zgcd a b)).
rewrite Zopp_involutive => X.
rewrite Zgcd_sym; apply Zis_gcd_gcd; by [apply Zgcd_is_pos | apply X, Zgcd_is_gcd].
Qed.

Lemma Zgcd_even : forall a b, Zeven a -> Zeven b -> Zgcd a b = 2 * Zgcd (a / 2) (b / 2).
Proof.
move=> a b Ha Hb; apply Zis_gcd_gcd.
- apply Zmult_le_0_compat => //; by apply Zgcd_is_pos.
- case/Zeven_ex: Ha => qa Ha; subst a.
  case/Zeven_ex: Hb => qb Hb; subst b.
  rewrite {2}(Zmult_comm 2 qa) {2}(Zmult_comm 2 qb) !Z_div_mult //.
  by apply Zis_gcd_mult, Zgcd_is_gcd.
Qed.

Lemma Zgcd_mult : forall a b k, 0 <= k -> Zgcd (k * a)  (k * b) = k * Zgcd a b.
Proof.
move=> a b k Hk.
apply Zis_gcd_gcd => //.
apply Zmult_le_0_compat => //; by apply Zgcd_is_pos.
by apply Zis_gcd_mult, Zgcd_is_gcd.
Qed.

Lemma Zgcd_even_odd : forall a b, Zeven a -> Zodd b -> Zgcd a b = Zgcd (a / 2) b.
Proof.
move=> a b Ha Hb; apply Zis_gcd_gcd; first by apply Zgcd_is_pos.
case/Zeven_ex: Ha => qa Ha; subst a.
case/Zodd_ex: Hb => qb Hb; subst b.
rewrite {2}(Zmult_comm 2) Z_div_mult_full //; by apply Zis_gcd_even_odd_new, Zgcd_is_gcd.
Qed.

Lemma Zgcd_Zpower_odd : forall k a b, Zodd b -> Zgcd (a * 2 ^^ k) b = Zgcd a b.
Proof.
elim=> [a b Hb /=|n IH a b Hb]; first by rewrite Zmult_1_r.
rewrite ZpowerS (Zmult_comm 2) Zmult_assoc Zgcd_even_odd //.
by rewrite Z_div_mult_full // IH.
by apply Zeven_mult_Zeven_r.
Qed.

Lemma Zgcd_for_euclid a b q : Zgcd (a + q * b) b = Zgcd a b.
Proof.
move: (Zis_gcd_for_euclid2 b (Zgcd a b) q a) => X.
apply Zis_gcd_gcd; first by apply Zgcd_is_pos.
rewrite Zplus_comm Zmult_comm; by apply Zis_gcd_sym, X, Zgcd_is_gcd.
Qed.

(** Zbeta *)

Definition Zbeta' e := Zpower 2 (e * 32).
Definition Zbeta := nosimpl Zbeta'.
Notation "\B^ n" := (Zbeta n) (at level 4, format "\B^ n") : zarith_ext_scope.

Lemma ZbetaE : forall e, \B^ e = Zpower 2 (e * 32).
Proof. done. Qed.

Lemma Zbeta1_Zpower2 : \B^ 1 = 2 ^^ 32. Proof. done. Qed.

Lemma Zpower_64_Zbeta : 2 ^^ 64 = Zbeta 2. Proof. done. Qed.

Lemma Zbeta_is_exp n m : Zbeta (n + m) = Zbeta n * Zbeta m.
Proof. by rewrite ZbetaE mulnDl ZpowerD. Qed.

Lemma Zbeta_S n : \B^n.+1 = \B^1 * \B^n.
Proof. by rewrite -Zbeta_is_exp. Qed.

Lemma Zbeta_0 l : 0 < \B^l.
Proof. by apply Zpower_0. Qed.

Lemma Zbeta_0' n : 0 <= \B^n.
Proof. by apply/Zlt_le_weak/Zbeta_0. Qed.

Lemma Zbeta_1 l : 1 <= \B^l.
Proof. by apply Zpower_1. Qed.

Lemma Zbeta_lt m n : (m < n)%nat -> \B^ m < \B^ n.
Proof. move=> H; rewrite /Zbeta; apply Zpower_2_lt => //; by rewrite ltn_mul2r. Qed.

Lemma Zbeta_le m n : (m <= n)%nat -> \B^m <= \B^n.
Proof. by move=> H; apply/ZleP; rewrite /Zbeta Zpower_2_le // leq_pmul2r. Qed.

Lemma Zpower_shift_2 n : 4 * n < \B^1 -> n < 2 ^^ 30.
Proof.
rewrite Zbeta1_Zpower2.
have -> : (32 = 2 + 30)%nat by rewrite //=.
rewrite ZpowerD [Zmult]lock /= -lock; omega.
Qed.

Lemma Zbeta1_1 a : ~ a * \B^1 = 1.
Proof.
move=> H.
move: (Ztrichotomy a 0) => [X | [X | X]].
- by destruct a.
- by rewrite X in H.
- suff H' : 1 < a * \B^1 by omega.
  apply Zle_lt_trans with (a * 1).
  + rewrite Zmult_1_r; omega.
  + by apply Zmult_gt_0_lt_compat_l.
Qed.

Lemma inv_mod_beta0 m : 1 * m + 0 * \B^1 = m. Proof. ring. Qed.
Lemma inv_mod_beta1 m : 0 * m + 1 * \B^1 = \B^1. Proof. ring. Qed.
Lemma inv_mod_beta2 : forall m d, Zis_gcd m (\B^1) d -> Zis_gcd m (\B^1) d. Proof. done. Qed.

Definition inv_mod_beta m :=
  match euclid_rec m \B^1 \B^1 (Zbeta_0' 1) 1 0 m 0 1 (inv_mod_beta0 m) (inv_mod_beta1 m) (inv_mod_beta2 m) with
    Euclid_intro u _ d _ _ => 
    if d <=? 0 then u mod (\B^1) else 
      if Zeq_bool u 0 then 0 else 
        - (u mod (\B^1)) + \B^1
  end.

(** Equality modulo *)

Definition eqmod (a b m : Z) := exists k : Z, a = b + k * m.

Notation "a =m b {{ m }}" := (eqmod a b m) (at level 80) : eqmod_scope.
Local Open Scope eqmod_scope.

Lemma eqmod_Zmod a b m : 0 < m -> (a =m b {{ m }} <-> a mod m = b mod m).
Proof.
move=> Hm; split.
- move=> [k' ->]; by rewrite Z_mod_plus_full.
- move=> H.
  rewrite (Z_div_mod_eq a m (Zlt_gt _ _ Hm)) (Z_div_mod_eq b m (Zlt_gt _ _ Hm)) H.
  exists ((a / m) - (b / m)); ring.
Qed.

Lemma eqmod_Zmod2 a b m : 0 <= m -> (a =m b {{ m }} -> a mod m = b mod m).
Proof. move=> Hm [k' ->]; by rewrite Z_mod_plus_full. Qed.

Lemma eqmod_refl : forall a m, a =m a {{ m }}. Proof. exists 0 => /=; ring. Qed.

Lemma eqmod_sym a b m : a =m b {{ m }} -> b =m a {{ m }}.
Proof. move=> [k ->]; exists (- k); ring. Qed.

Lemma eqmod_trans a b c m : a =m b {{ m }} -> b =m c {{ m }} -> a =m c {{ m }}.
Proof.
rewrite /eqmod; move=> [kab Hab] [kbc Hbc].
exists (kab + kbc).
rewrite Hab Hbc; ring.
Qed.

Lemma eqmod_compat_plus_R a b c m : a =m c {{ m }} -> a + b =m c + b {{ m }}.
Proof. move=> [kac Hac]. exists kac. rewrite Hac; ring. Qed.

Lemma eqmod_div k m a b : a =m b {{ k * m }} -> a =m b {{ m }}.
Proof. move=> [x Hx]; exists (x * k); rewrite Hx; ring. Qed.

(* TODO: change *)
Lemma eqmod_reg_mult a b d m : a =m b {{ m }} -> a * d =m b * d {{ m }}.
Proof. move=> [k' ->]. exists (k' * d); ring. Qed.

(* TODO: change *)
Lemma eqmod_reg_mult_l a b d m : a =m b {{ m }} -> d * a =m d * b {{ m }}.
Proof. move=> [k' ->]. exists (k' * d); ring. Qed.

Lemma eqmod_rewrite m a a' b c : a =m a' {{ m }} ->  a * b =m c {{ m }} -> a' * b =m c {{ m }}.
Proof.
move=> [ka Hka] [kab Hkab].
rewrite Hka {Hka} in Hkab.
ring_simplify in Hkab.
exists (kab - ka * b).
apply Zplus_minus_eq in Hkab.
rewrite Hkab; ring.
Qed.

Lemma eqmod_reg_mult' a b d m : Zis_gcd m d 1 ->
  a * d =m b * d {{ m }} -> a =m b {{ m }}.
Proof.
move=> Hgcd H.
move: (euclid m d).
inversion 1 as [m' d' gcd Hdd' Hgcd'].
case: {Hgcd Hgcd'}(Zis_gcd_uniqueness_apart_sign _ _ _ _ Hgcd Hgcd') => Hinv; subst gcd.
- have Hinv' : d * d' =m 1 {{ m }}.
    exists (-m'); rewrite Hinv; ring.
  move: {H}(eqmod_reg_mult _ _ d' _ H) => H.
  rewrite -Zmult_assoc Zmult_comm in H.
  move: {H}(eqmod_rewrite _ _ _ _ _ Hinv' H).
  rewrite Zmult_1_l.
  move/eqmod_sym.
  rewrite -Zmult_assoc Zmult_comm => H.
  move/(eqmod_rewrite _ _ _ _ _ Hinv') : H => H.
  rewrite Zmult_1_l in H.
  by apply eqmod_sym.
- have Hinv' : d * d' =m -1 {{ m }}.
    exists (- m').
    have X : 1 + d * d' = - m' * m by rewrite Hinv; ring.
    omega.
  move: {H}(eqmod_reg_mult _ _ d' _ H) => H.
  rewrite -Zmult_assoc Zmult_comm in H.
  move: {H}(eqmod_rewrite _ _ _ _ _ Hinv' H) => H.
  apply eqmod_sym in H.
  rewrite -Zmult_assoc Zmult_comm in H.
  move: {H}(eqmod_rewrite _ _ _ _ _ Hinv' H) => [k H].
  exists k.
  have X : k * m = -1 * b + 1 * a by omega.
  rewrite X; ring.
Qed.

Lemma eqmod_reg_mult_beta_odd a b k m : Zodd m ->
  \B^k * a =m \B^k * b {{ m }} -> a =m b {{ m }}.
Proof.
move=> Hm.
rewrite Zmult_comm.
move/eqmod_sym.
rewrite Zmult_comm => H.
apply eqmod_reg_mult' in H.
apply eqmod_sym => //.
rewrite ZbetaE.
by apply/Zis_gcd_sym/Zis_gcd_Zpower_odd.
Qed.

Lemma eqmod_minmod m a b k : a =m b {{ m }} -> a - k * m =m b {{ m }}.
Proof. move=> [k' H]. exists (k' - k); rewrite H; ring. Qed.

Lemma eqmod_mod m a b : 0 < m -> a =m b {{ m }} -> a =m b mod m {{ m }}.
Proof. move=> Hm => /eqmod_Zmod; rewrite eqmod_Zmod // Zmod_mod; by auto. Qed.

Lemma eqmod_inv a a' m : 0 <= a < m -> 0 <= a' < m -> a =m a' {{ m }} -> a = a'.
Proof.
move=> [Ham1 Ham2] [Ha'm1 Ha'm2] [[|p|p] H].
- by rewrite Zmult_0_l Zplus_0_r in H.
- have Htmp : m <= a.
    rewrite H -{1}(Zplus_0_l m).
    apply Zplus_le_compat => //.
    rewrite -{1}(Zmult_1_l m).
    apply Zmult_gt_0_le_compat_r => //.
    omega.
    by destruct p.
  omega.
- have Htmp : a < 0.
    rewrite H.
    have : forall a b, a < -b -> a + b < 0 by move=> a0 b; omega.
    apply.
    rewrite Zopp_mult_distr_l Zopp_neg.
    apply Zlt_le_trans with m => //.
    rewrite -{1}(Zmult_1_l m).
    apply Zmult_le_compat_r.
    by destruct p.
    omega.
  omega.
Qed.

Local Close Scope eqmod_scope.
Local Close Scope zarith_ext_scope.

Fixpoint isum n := if n is n'.+1 then ((isum n') + n'.+1)%nat else O. 

Definition Zisum z := if z is Zpos p then Z_of_nat (isum (nat_of_P p)) else 0.

Lemma Zisum_prop' : forall a, 0 <= a -> Zisum (a + 1) = Zisum a + (a + 1).
Proof.
case=> // p.
move=> _.
rewrite /= nat_of_P_plus_morphism plus_comm /=.
rewrite inj_plus /= -Z_of_nat_Zpos_P_of_succ_nat inj_plus.
by rewrite Zpos_plus_distr (Zpos_eq_Z_of_nat_o_nat_of_P p).
Qed.

(** factorial in Z *)

Definition Zfact (z : Z) : Z :=
  match z with
    | Zpos p => Z_of_nat (fact (nat_of_P p))
    | _ => 1
  end.

Lemma Zfact_neg : forall z, z < 0 -> Zfact z = 1.
Proof. by destruct z. Qed.

Lemma Zfact_zero : Zfact 0 = 1. Proof. done. Qed. 

Lemma fact_lemma : forall n, (0 < n)%nat -> (fact n = n * fact (n - 1))%nat.
Proof.
elim=> //; case=> [ // |n] IH _.
by rewrite subSS subn0 IH //= subSS subn0 !mulnE.
Qed.

Lemma Zfact_pos : forall z, z > 0 -> Zfact z = z * Zfact (z - 1).
Proof.
case => // p Hp.
rewrite [X in X = _]/=.
rewrite fact_lemma; last by apply/ltP/Pos2Nat.is_pos.
rewrite inj_mult positive_nat_Z.
f_equal.
destruct p => //.
  rewrite (_ : Z.pos p~1 - 1 = Z.pos p~0) //=.
  do 2 f_equal.
  by rewrite Pos2Nat.inj_xO Pos2Nat.inj_xI subSS subn0.
rewrite /=.
do 2 f_equal.
by rewrite -Pos.succ_pred_double Pos2Nat.inj_succ subSS subn0.
Qed.

Local Open Scope zarith_ext_scope.

Fixpoint bbs_fun_rec (n : nat) (seed m : Z) {struct n} : seq bool :=
  match n with
    | O => nil
    | n.+1 => (Zodd_bool seed) :: (bbs_fun_rec n (seed ^^ 2 mod m) m)
  end.

Definition bbs_fun (len:nat)(seed m:Z) : seq bool :=
  bbs_fun_rec len ((seed * seed) mod m) m.

Lemma bbs_fun_rec_cat0 : forall n seed modu,
  bbs_fun_rec n.+1 (seed mod modu) modu =
  bbs_fun_rec n (seed mod modu) modu ++ 
  bbs_fun_rec 1 ((seed mod modu)^^(2 ^ n) mod modu) modu.
Proof.
elim => [seed modu | n0 IHn0 seed modu].
- by rewrite /= Zmult_1_r Zmod_mod.
- by rewrite [n0.+1]lock /= -lock IHn0 /= Zmult_1_r expnS 
    {2}Zpower_b_square (Zpower_exp_mod (2 ^ n0) 2 (seed mod modu) modu).
Qed.

Lemma bbs_fun_rec_cat0' : forall n seed modu,
  bbs_fun_rec n.+1 (seed mod modu) modu =
  bbs_fun_rec 1 (seed mod modu) modu ++ 
  bbs_fun_rec n ((seed mod modu)^^2 mod modu) modu.
Proof. by elim. Qed.

Lemma bbs_fun_rec_cat1 : forall n k seed modu,
  bbs_fun_rec (n + k) (seed mod modu) modu =
  bbs_fun_rec n (seed mod modu) modu ++ 
  bbs_fun_rec k ((seed mod modu) ^^ (2 ^ n) mod modu) modu.
Proof.
elim => [k seed modu | n0 IHn0 k seed modu].
- by rewrite /= Zmult_1_r Zmod_mod.
- by rewrite addSnnS IHn0 -{1}(addn1 n0) IHn0 -catA 
   bbs_fun_rec_cat0' Zpower_exp_mod expnS mulnC.
Qed.

Lemma bbs_fun_rec_cat : forall k n seed modu,
  bbs_fun_rec (k * n.+1) (seed mod modu) modu =
  bbs_fun_rec (k * n) (seed mod modu) modu ++ 
  bbs_fun_rec k ((seed mod modu)^^(2 ^ (k * n)) mod modu) modu.
Proof.
elim=> // k IHk n0 seed modu.
have -> : (k.+1 * n0.+1 = k.+1 * n0 + k.+1)%nat by rewrite mulnS addnC. 
by rewrite bbs_fun_rec_cat1.
Qed.
