(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
(* begin hide *)
From mathcomp Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq.
From mathcomp Require Import path.

Set Implicit Arguments.
Unset Strict Implicit.
(* end hide *)

Section Type_ext.

Variable A : Type.

Lemma path_rel_inj: forall (rel1: rel A) rel2,
  (forall from,
    forall to, rel1 from to = rel2 from to) ->
  forall tl hd,
    path rel1 hd tl = path rel2 hd tl.
Proof.
move => rel1 rel2 Heq.
by elim => //= hd' tl -> hd; rewrite Heq.
Qed.
  
Lemma cycle_rel_inj: forall (rel1: rel A) rel2,
  (forall from,
    forall to, rel1 from to = rel2 from to) ->
  forall l,
    cycle rel1 l = cycle rel2 l.
Proof.
move => rel1 rel2 Heq [] //= hd tl.
by apply path_rel_inj.
Qed.

End Type_ext.

Section eqType_ext.

Variable A : eqType.

Lemma ucycle_rel_inj: forall (rel1: rel A) rel2,
  (forall from,
    forall to, rel1 from to = rel2 from to) ->
  forall l,
    ucycle rel1 l = ucycle rel2 l.
Proof.
move => rel1 rel2 Heq [] //= hd tl.
rewrite /ucycle.
by rewrite (@cycle_rel_inj _ rel1 rel2).
Qed.

Lemma path_R_behead: forall (P: pred A) (R: rel A),
  (forall x y, R x y -> P y) ->
  forall (tl: seq A) (hd: A),
    path R hd tl ->
    forall (x: A), x \in tl ->
      P x.
Proof.
move => P R HPR.
elim => //= t s Hind hd /andP [] HhdRt Hpath x.
rewrite in_cons => [] /orP [].
- case/eqP => ->.
  exact: HPR _ _ HhdRt.
- move => Hin.  
  apply: (Hind t) => //.
Qed.  

Lemma path_R_belast: forall (P: pred A) (R: rel A),
  (forall x y, R x y -> P x) ->
  forall (hd: A) (tl: seq A),
    path R hd tl ->
    forall (x: A), x \in belast hd tl ->
      P x.
Proof.
move => P R HPR hd.
apply: last_ind => //=.
move => s x Hs.
rewrite rcons_path => /andP [] HPaths Hlast.
move: {Hs}(Hs HPaths) => Hin.
move => i.
rewrite belast_rcons in_cons => /orP //= [].
- case /eqP => ->.
  case: s HPaths Hlast Hin => //= [ _ HhdRx _ | hd' tl] .
  + exact: HPR HhdRx.
  + move => /andP [] HhdRhd' _ _ _.
    exact: HPR HhdRhd'.
- case: (lastP s) Hlast Hin => //= s0 x0.
  rewrite last_rcons => HR Hbelast .  
  rewrite mem_rcons in_cons => /orP [].
  - case /eqP => ->.
    exact: (HPR _ _ HR).
  - move => Hin.
    apply: Hbelast.
    rewrite belast_rcons in_cons.
    apply /orP.
    by apply or_intror.
Qed.

End eqType_ext.



