(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrfun ssrbool eqtype ssrnat.
Require Import ZArith.

Local Open Scope Z_scope.

Lemma ZeqP : Equality.axiom Zeq_bool.
Proof. move=> x y; apply: (iffP idP) => H; by apply/Zeq_is_eq_bool. Qed.
  
Canonical Z_eqMixin := EqMixin ZeqP.
Canonical Z_eqType := Eval hnf in EqType Z Z_eqMixin.

Implicit Arguments ZeqP [x y].
Prenex Implicits ZeqP.

Notation "'Z<=nat'" := (Z.of_nat) (at level 9) : zarith_ext_scope.
Open Local Scope zarith_ext_scope.

(** ssreflect-like naming for ZArith *)

Definition Zaddn0 := Zplus_0_r.
Definition Zadd0n := Zplus_0_l.

Definition ZaddnC : ssrfun.commutative Zplus := Zplus_comm.
Definition ZaddnA : ssrfun.associative Zplus := Zplus_assoc.

(* Z.leb_spec0 : forall x y : Z, Bool.reflect (x <= y) (x <=? y) *)
Lemma ZleP m n : reflect (m <= n) (Zle_bool m n).
Proof. apply: (iffP idP); by apply Z.leb_le. Qed.

Implicit Arguments ZleP [m n].
Prenex Implicits ZleP.

(* Z.ltb_spec0  forall x y : Z, reflect (x < y) (x <? y) *)
Lemma ZltP m n : reflect (m < n) (m <? n).
Proof. apply: (iffP idP); by apply Z.ltb_lt. Qed.

Implicit Arguments ZltP [m n].
Prenex Implicits ZltP.

Lemma ZgeP m n : reflect (m >= n) (m >=? n).
Proof. apply: (iffP idP); rewrite /Zge /Zge_bool; by destruct (m ?= n). Qed.

Implicit Arguments ZgeP [m n].
Prenex Implicits ZgeP.

Lemma ZgtP m n : reflect (m > n) (m >? n).
Proof. apply: (iffP idP); rewrite /Zgt /Zgt_bool; by destruct (m ?= n). Qed.

Implicit Arguments ZgtP [m n].
Prenex Implicits ZgtP.

Definition ZleqNgt := Z.leb_antisym.
Definition ZltnNge := Z.ltb_antisym.

Definition Zltnn := Z.ltb_irrefl.

Lemma ZltnW (m n : Z) : m <? n -> m <=? n.
Proof. move/ZltP => ?; apply/ZleP; omega. Qed.

Lemma Zltn_neqAle (m n : Z) : (m <? n) = (m != n) && (m <=? n).
Proof.
apply/idP/idP.
  move/ZltP => mn.
  apply/andP; split.
    apply/eqP; omega.
  by apply/ZltnW/ZltP.
case/andP => /eqP H1 /ZleP H2; apply/ZltP; omega.
Qed.

Lemma Zleq_add2r (p m n : Z) : (m + p <=? n + p) = (m <=? n).
Proof.
move Hlhs : (_ + _ <=? _) => [].
  by apply/esym/ZleP/(Zplus_le_reg_r _ _ p)/ZleP.
apply/esym/negP.
move/negP in Hlhs.
contradict Hlhs.
by apply/ZleP/Zplus_le_compat_r/ZleP.
Qed.

Lemma Zltn_add2r (p m n : Z) : (m + p <? n + p) = (m <? n).
Proof.
move Hlhs : (_ + _ <? _) => [].
  by apply/esym/ZltP/(Zplus_lt_reg_r _ _ p)/ZltP.
apply/esym/negP.
move/negP in Hlhs.
contradict Hlhs.
by apply/ZltP/Zplus_lt_compat_r/ZltP.
Qed.

Lemma Zltn_add2l (p m n : Z) : (p + m  <? p + n) = (m <? n).
Proof. by rewrite 2!(ZaddnC p) Zltn_add2r. Qed.

Lemma Zleq_sub2r n m p : (n <= m -> n - p <= m - p)%Z.
Proof. move=> H; omega. Qed.

Lemma Zltn_sub2r n m p : n < m -> n - p < m - p.
Proof. move=> H; omega. Qed.

Definition Zmul1n : ssrfun.left_id 1%Z Z.mul := Zmult_1_l.

Definition Zmuln1 : ssrfun.right_id 1%Z Z.mul := Zmult_1_r.

Definition ZmulnC : ssrfun.commutative Zmult := Zmult_comm.

Definition ZmulnDl := Z.mul_add_distr_r.

Definition ZmulnDr := Z.mul_add_distr_l.

Definition ZmulnA : ssrfun.associative Zmult := Zmult_assoc.

Lemma Zleq_pmul2r m n1 n2 : (0 < m)%Z -> (n1 * m <=? n2 * m = (n1 <=? n2))%Z.
Proof.
move=> Hm.
move Hlhs : (_ * _ <=? _) => [].
  move/ZleP in Hlhs.
  apply Zmult_le_reg_r in Hlhs; last by apply Zlt_gt.
  by move/ZleP in Hlhs.
apply/esym/negbTE.
move/negbT : Hlhs.
apply contra => Hlhs.
apply/ZleP/Zmult_le_compat_r; last by apply Z.lt_le_incl.
by apply/ZleP.
Qed.

Lemma Zleq_pmul2l m n1 n2 : (0 < m)%Z -> (m * n1 <=? m * n2 = (n1 <=? n2))%Z.
Proof. move=> ?; by rewrite -(ZmulnC n1) -(ZmulnC n2) Zleq_pmul2r. Qed.

Lemma Zltn_pmul2r m n1 n2 : (0 < m)%Z -> (n1 * m <? n2 * m = (n1 <? n2))%Z.
Proof.
move=> Hm.
move Hlhs : (_ * _ <? _)%Z => [].
  move/ZltP/Zmult_lt_reg_r/ZltP in Hlhs.
  by rewrite Hlhs.
apply/esym/negbTE.
move/negbT : Hlhs.
apply contra => Hlhs.
apply/ZltP/Zmult_lt_compat_r => //; by apply/ZltP.
Qed.

(* other *)

Lemma Z_of_nat_le n m : (n <= m)%nat = (Z<=nat n <=? Z<=nat m)%Z.
Proof.
move Hlhs : (n <= m)%nat => [].
  by apply/esym/ZleP/Nat2Z.inj_le/leP.
apply/esym/negbTE.
move/negbT : Hlhs.
apply: contra => Hlhs.
by move/ZleP/Nat2Z.inj_le/leP in Hlhs.
Qed.

Lemma Z_of_nat_lt n m : (n < m)%nat = (Z<=nat n <? Z<=nat m)%Z.
Proof.
move Hlhs : (n < m)%nat => [].
  by apply/esym/ZltP/Nat2Z.inj_lt/ltP.
apply/esym/negbTE.
move/negbT : Hlhs.
apply contra => Hlhs.
by move/ZltP/Nat2Z.inj_lt/ltP in Hlhs.
Qed.
