(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
Require Import Omega.
From mathcomp Require Import ssreflect ssrbool eqtype ssrnat.
Export ssrnat.

Lemma leq_and_ltn' l l' : l <= l' < l = false.  
Proof.
apply/andP.
case => H1 H2.
have := leq_ltn_trans H1 H2.
by rewrite ltnn.
Qed.

Lemma leq_and_ltn'' l l' : l <= l' < l.+1 -> l = l'.
Proof.
case/andP => [H1 H2].
rewrite leq_eqVlt in H1.
case/orP: H1 => [/eqP // | H1].
have := leq_ltn_trans H1 H2.
by rewrite ltnn.
Qed.

Lemma ltn_total l l' : l < l' \/ l' <= l.
Proof.
case/boolP : (l == l') => [/eqP -> | ].
  right; by rewrite leqnn.
rewrite neq_ltn.
case/orP; auto.
Qed.
  
Lemma ltn_leq_trans n m p : m < n -> n <= p -> m < p.
Proof. move=> Hmn; exact: leq_trans. Qed.

Lemma ltn_leq_add2l : forall n1 n2 m1 m2,
  n1 <= n2 -> m1 < m2 -> n1 + m1 < n2 + m2.
Proof.
move=> n1 [|n2] m1 m2.
by case: n1.
move=> *; by rewrite -addnS leq_add.  
Qed.

Lemma ltn_sub_add : forall m n p, 0 < p -> (m - n < p) = (m < n + p).
Proof. move=> m n [//|p] _; by rewrite ltnS leq_subLR addnS. Qed.

Lemma eq_succ : forall l, ~ l = l.+1.
Proof. by elim. Qed.

Lemma nat_rec_inv ub (P : nat -> Prop) :
  P ub ->
  (forall n, P n.+1 -> n < ub -> P n) ->
  forall i, i <= ub -> 
  P i.
Proof.
elim: ub => //=.
- move => P0 _ i.
  by rewrite leqn0 => /eqP ->.
- move => n Hind HPSn HPsub i0.
  rewrite leq_eqVlt.
  case/orP => /= [ /eqP -> // | ].
  rewrite ltnS leq_eqVlt.
  case/orP => [/eqP -> | Hi0].
  + by apply HPsub.
  + apply Hind => //=.
    * by apply HPsub.
    * move => n0 HPSn0 Hn0.
      apply HPsub => //=.
      by rewrite (ltn_trans Hn0).
    * by rewrite ltnW.
Qed.

Ltac ssromega :=
  (repeat arith_hypo_ssrnat2coqnat ;
  arith_goal_ssrnat2coqnat ; 
  omega)
with arith_hypo_ssrnat2coqnat :=
  match goal with
    | H : context [?L < ?R] |- _ 	=> move/ltP: H => H
    | H : context [?L <= ?R] |- _ 	=> move/leP: H => H
    | H : context [?L < ?M < ?R] |- _ 	=> let H1 := fresh in case/andP: H => H H1
    | H : context [?L <= ?M < ?R] |- _ 	=> let H1 := fresh in case/andP: H => H H1
    | H : context [?L <= ?M <= ?R] |- _ 	=> let H1 := fresh in case/andP: H => H H1
    | H : context [addn ?L ?R] |- _ => rewrite <- plusE in H
    | H : context [muln ?L ?R] |- _ => rewrite <- multE in H
    | H : context [subn ?L ?R] |- _ => rewrite <- minusE in H
(*    | H : ?a <> ?b |- _ => 
      match type of a with
      | nat => apply Coq.Arith.Compare_dec.not_eq in H
      end*)
  end
with arith_goal_ssrnat2coqnat := 
  rewrite -?plusE -?minusE -?multE;
  match goal with
    | |- is_true (_ < _)%nat => apply/ltP
    | |- is_true (_ <= _)%nat => apply/leP
    | |- is_true (_ && _) => apply/andP; split; ssromega
    | |- _ /\ _ => split; ssromega
    | |- _ \/ _ => (left; ssromega) || (right; ssromega)
    | |- _ => idtac  
  end.

Fact increasing :
  forall (m : nat) (A : nat -> nat),
    (forall m : nat, A m <= A (m.+1)) ->
    forall n, A m <= A (m + n).
Proof.
move => m A inc. elim => [| n].
- rewrite addn0; done.
- have -> : m + succn n = succn (m + n) by rewrite -addnS; done.
  simpl.
  have : A (m + n) <= (A (m + n).+1) by exact: inc.
  move => H K.
  move: K H; exact: leq_trans.
Qed.

Lemma bounded_inc_convergent : forall max
    (A : nat -> nat),
    (forall n, A n.+1 >= A n) ->
      (forall m, m <= max.+1 -> A m <= max) ->
      (forall K L,
         A K = L -> A K.+1 = L ->
         forall m : nat, K <= m -> A m = L) ->
      forall m, max <= m -> A max = A m.
Proof.
elim.
- move => A A_inc H same_stays.
  have -> : A 0 = 0 by apply/eqP; rewrite -leqn0 ; exact: H.
  move => m mlarge; symmetry; apply: (same_stays O) => //; apply/eqP; rewrite -leqn0; by apply H.
- move => max IH A A_inc small same_stays.
  case_eq (A 1) => a1.
  - have H : forall m : nat, 0 <= m -> A m = 0.
    - apply: same_stays; [ | assumption].
      apply/eqP; rewrite -leqn0 -{2}a1.
      apply: increasing; exact: A_inc.
    move => m _; rewrite !H; done.
  - move => a1_; pose B n := (A n.+1).-1.
    move => m max_m.
    have -> : A m = (B (m.-1)).+1.
    - rewrite/B; have -> : succn (predn m) = m by
      move: max_m; exact: ltn_predK.
      rewrite prednK; [done | ].
      move: max_m.
      case: m => [ | m]; [done | ].
      move => _.
      elim: m => [ | m]; [rewrite a1_; done | ].
      have H : A (succn m) <= A (succn (succn m)) by exact: A_inc.
      rewrite !ltnNge.
      apply: contra; apply: leq_trans; assumption.
    have -> : A (succn max) = (B max).+1.
    - rewrite/B prednK; [done |].
      move: IH max_m => _ _; move: small => _.
      have H : A 1 <= A (succn max).
      - have -> : succn max = 1 + max by done.
        apply: increasing; by eapply A_inc.
      rewrite ltnNge.
      apply/negP.
      move => I.
      have : A 1 <= 0 by move: H I; exact: leq_trans.
      rewrite a1_; done.
    f_equal.
    apply: IH.
    - rewrite/B.
      move => n.
      have H : n = predn (succn n) by done.
      rewrite -!subn1; apply: leq_sub2r; exact: A_inc.
    - case => [ | n].
      + move => _.
        rewrite/B.
        have ? : A 1 <= succn max by apply: small; done.
        rewrite - (leq_add2r 1).
        have -> : predn (A 1) + 1 = A 1 by
          rewrite addn1 prednK; [ | rewrite a1_]; done.
        rewrite addn1; assumption.
      + rewrite/B => nsmall.
        suff H : (A (succn (succn n))) <= max.+1.
        - move: H. set a := A (succn (succn n)).
          rewrite -subn1 -addn1 => H.
          move: (leq_sub2r 1 H) => I.
          suff -> : max = max + 1 - 1 by assumption.
          rewrite addn1 subn1; done.
        apply: small.
        rewrite -(ltn_add2r 1) in nsmall.
        have -> : succn n = n + 1 by rewrite addn1; done.
        have -> : succn (succn max) = succn max + 1 by rewrite addn1; done.
        assumption.
    - move => K L; rewrite/B => KL KKL x Kx; apply/eqP.
      rewrite -(eqn_add2r 1).
      have -> : predn (A (succn x)) + 1 = A (succn x).
      * rewrite addn1; apply: prednK.
        suff H : succn a1 <= A (succn x) by apply: leq_ltn_trans H.
        rewrite -a1_; apply: increasing; exact: A_inc.
      apply/eqP; apply: (same_stays (succn K)).
      - move/eqP: KL; rewrite -(eqn_add2r 1); move/eqP => KL.
        rewrite -KL addn1 prednK;  [reflexivity | ].
        suff I : succn a1 <= A (succn K) by apply: leq_ltn_trans I.
        rewrite -a1_.
        apply: increasing; exact: A_inc.
      - move/eqP: KKL; rewrite -(eqn_add2r 1); move/eqP => KKL.
        rewrite - KKL addn1 prednK; [reflexivity | ].
        suff I : succn a1 <= A (succn (succn K)) by apply: leq_ltn_trans I.
        rewrite -a1_; apply: increasing; exact: A_inc.
      - exact: Kx.
    - rewrite -ltnS; by apply: ltn_leq_trans (leqSpred m).
Qed.